"""Behavior, Query, and State
=============================

The decision and control part is structured as a directed, possibly
cyclic graph. The following describes `Behavior`, `Query`, and `State`
(BQS) classes. TL;DR: transfer between states of FSM is done by BT.

`Behavior` nodes correspond to *control flow* nodes of Behavior Tree
(BT). `Query` nodes correspond to *condition* nodes of BT. `State` nodes
correspond to root of BT, *action* nodes of BT, as well as states of
Finite State Machine (FSM). BQS differences to BT and FSM are discussed
bellow.

`Behavior` nodes are non-leafy nodes with the member `children`, which
is a list of other BQS nodes. The `children` nodes are processed in an
order corresponding to the behavior of the node subclass. The return
value of the `decide()` method is also specific to the node subclass.

Examples of `Behavior` nodes are `Sequence`, `Fallback`, and `Not`.

`Query` nodes are leaf nodes that query the `world` in read-only
fashion. The decision process cannot stop at the `Query` node.

Examples of `Query` nodes are `AlwaysTrue` and `AlwaysFalse`.

`State` nodes are leaf nodes where the decision process begins and ends.
Every `State` node is the root of BT and must have had some `Behavior`
node assigned to the ``change_state`` member.

`Behavior` and `Query` nodes must implement ``_decide()`` method.

`State` nodes are expected to implement ``_control()`` method.


Inheritance diagram
-------------------

.. inheritance-diagram:: pdc.bqs


Comparison to FSM
-----------------

Compared to FSM, BQS has framework for transition between states.
``change_state`` is the root of the BT, the destination states are
`State` leaves of that BT.


Comparison to BT
----------------

BQS can simulate BT when only one behavior is specified and assigned to
``change_state`` member of every `State`.

Each call to `decide()` always returns some object or ``False``. The
*running* state known from BT is not possible because BT is only used to
change between (FSM) states. (To change between BT actions.)


World
-----

Each BQS node has `world` member. `World` is shared between BQS nodes
and contain the perceived information, the history, and the status (of
the world).

It is possible to define multiple worlds and then assign different BQS
nodes to different worlds. To share information between worlds, the
`State` nodes of the first world can implement method that updates the
second world, i.e., the ``_control(world)`` method.


Members
-------

"""
# License
# -------
#
# Copyright (c) 2022 Jiri Vlasak
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


class World:
    """History and status of the world."""
    def __init__(self, history_len=2):
        self.history = [{} for i in range(history_len)]
        self.status = {
            "deciding": False,
            "last_state": None,
            "read_world": 0,
            "write_world": 0}

    def cs(self):
        """Return current (last set) `State`."""
        return self.status["last_state"]

    def cw(self, element=None):
        """Return (element of) current (last update of) world."""
        try:
            return self.history[self.status["read_world"]][element]
        except KeyError:
            return self.history[self.status["read_world"]]

    def include(self, *nodes):
        """Include BQS nodes into the self (world).

        :param nodes: Subclasses of BQS `Node`.
        """
        for n in nodes:
            assert isinstance(n, Node)
            n.world = self

    def _increase_write_world_index(self):
        """Begin addition of new world. Must be atomic."""
        assert self.status["read_world"] == self.status["write_world"]
        self.status["write_world"] += 1
        self.status["write_world"] %= len(self.history)

    def append_blank_world(self):
        """Add new blank world into the history. Must be atomic."""
        self._increase_write_world_index()
        self.history[self.status["write_world"]] = {}

    def append_copied_world(self):
        """Add copy of the last world into the history. Must be atomic."""
        self._increase_write_world_index()
        self.history[self.status["write_world"]] = dict(
            self.history[self.status["read_world"]])

    def append_new_world(self, **kwargs):
        """Add new world from keywords. Must be atomic."""
        self.append_blank_world()
        for k in kwargs:
            self.history[self.status["write_world"]][k] = kwargs[k]

    def world_added(self):
        """New world has been added, mark for read. Must be atomic."""
        assert self.status["read_world"] != self.status["write_world"]
        self.status["read_world"] = self.status["write_world"]


class Node:
    """Base node of BQS graph."""
    def decide(self, *args, **kwargs):
        """Make a decision, return obj or False. Calls `_decide()`."""
        print(self.__class__.__name__)
        ret = self._decide(*args, **kwargs)
        print("{} returns {}".format(self.__class__.__name__, ret))
        return ret

    def _decide(self):
        """Abstract. To be implemented by *Decision* subclasses."""
        raise NotImplementedError("Only for Decision nodes.")

    def control(self, *args, **kwargs):
        """Run a controller. Calls `_control()`."""
        print(self.__class__.__name__)
        ret = self._control(*args, **kwargs)
        print("{} returns {}".format(self.__class__.__name__, ret))
        return ret

    def _control(self):
        """Abstract. To be implemented by *Control* subclasses."""
        raise NotImplementedError("Only for Control nodes.")

    world = World()
    """Shared history of the world."""


class Behavior(Node):
    """Non-leaf decision node of BQS graph.

    No query to `world`. Just tick children.

    Subclasses must implement ``_decide()`` method.
    """
    def __init__(self, *children):
        for c in children:
            assert isinstance(c, Node)
        self.children = children
        """A list of `Node` instances to be evaluated."""


class Query(Node):
    """Read-only query to `world`.

    Leaf node of BQS, but can't end deciding here.

    Subclasses must implement ``_decide()`` method.
    """
    pass


class State(Node):
    """Leaf node where decision process begins and ends.

    Read-only query to `world`.

    Subclasses must have a `Behavior` assigned to ``change_state``
    member.

    Subclasses must implement ``_control()`` method.
    """
    def set_begin(self):
        """Set self as the beginning state."""
        self.world.status["last_state"] = self

    def _decide(self):
        """Transfer between states.

        Needs `Behavior` assigned to the ``change_state``. Return ``True`` if
        the `State` of the `world` has been changed and ``False`` otherwise.
        """
        if self.world.status["deciding"]:
            self.world.status["deciding"] = False
            self.world.status["last_state"] = self
        else:
            self.world.status["deciding"] = True
            self.change_state.decide()
            self.world.status["deciding"] = False
        return self.world.status["last_state"] is not self


# Behavior classes
class Sequence(Behavior):
    """-> in BT."""
    def _decide(self):
        """Return the last child's return if all children succeed or False."""
        ret = True
        ch = 0
        while (self.world.status["deciding"]
                and ch < len(self.children)
                and ret):
            ret = self.children[ch].decide()
            ch += 1
        return ret


class Fallback(Behavior):
    """? in BT."""
    def _decide(self):
        """Return the first successfull child's return or ``False``."""
        ret = False
        ch = 0
        while (self.world.status["deciding"]
                and ch < len(self.children)
                and not ret):
            ret = self.children[ch].decide()
            ch += 1
        return ret


class Not(Behavior):
    """Negation. For one child only."""
    def _decide(self):
        """Return negation of the first child."""
        return not self.children[0].decide()


# Query classes
class AlwaysTrue(Query):
    """Always return ``True``."""
    def _decide(self):
        """Return ``True``."""
        return True


class AlwaysFalse(Query):
    """Always return ``False``."""
    def _decide(self):
        """Return ``False``."""
        return False



"""Data Processing
==================

The processing of input data in the perception part is structured as a
directed acyclic graph. There may be multiple data source nodes, but
only one sink node. This approach allows parallel processing of data
from multiple sources. Intermediate nodes of the graph may use classes
defined in this module to process the data.


Inheritance diagram
-------------------

.. inheritance-diagram:: pdc.dp


Members
-------

"""
# License
# -------
#
# Copyright (c) 2022 Jiri Vlasak
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


def _get_float(w, i):
    if isinstance(w, list) or isinstance(w, tuple) or isinstance(w, dict):
        return float(w[i])
    else:
        return float(w)


class Filter:
    """Base class for the filters."""
    def update(self, curr):
        """Process curr, return processed.

        :param curr: Float or dict. Measured value.
        """
        print("{} filtering...".format(self.__class__.__name__))
        ret = self._update(curr)
        print("{} returns {}".format(self.__class__.__name__, ret))
        return ret

    def _update(self):
        """Abstract. To be implemented by `Filter` subclasses."""
        raise NotImplementedError("_update() is missing.")


class FiltersChain(Filter):
    """Chain multiple filters that are run in sequence.

    When any of the filters fails, the whole `FiltersChain` fails. To
    avoid problems of sharing filters between `FiltersChain` classes
    (it's Python) in `FiltersChain` constructor always call `Filter`
    constructor::

        FiltersChain(
            Filter(),
            Filter(),
            Filter())

    """
    def __init__(self, *filters):
        for f in filters:
            assert isinstance(f, Filter)
        self.filters = filters
        """A list of filters to be applied."""

    def _update(self, curr):
        """Run multiple filters on curr, stop if any of the filters fails."""
        for f in self.filters:
            curr = f.update(curr)
            if not curr:
                return curr
        return curr


class LowPassFilter(Filter):
    """See https://en.wikipedia.org/wiki/Low-pass_filter"""
    def __init__(self, weight=0.5, keys=[]):
        """Create new `LowPassFilter` object.

        :param weight: Float or list or dict of weights.
        :param keys: The list of keys if dict is provided to `_update()`.
        """
        self.last = None
        self.keys = keys
        if len(self.keys) > 0:
            self.weight = {}
            i = 0
            for k in self.keys:
                self.weight[k] = _get_float(weight, i)
                i += 1
        else:
            self.weight = _get_float(weight, None)

    def _update(self, curr):
        """Compute value of curr in Low-Pass filter fashion."""
        def f(c, w, la):
            return w * c + (1.0 - w) * la
        if self.last is None:
            self.last = curr
        elif isinstance(curr, dict):
            for k in curr:
                if k in self.keys:
                    self.last[k] = f(curr[k], self.weight[k], self.last[k])
                else:
                    self.last[k] = curr[k]
        else:
            self.last = f(curr, self.weight, self.last)
        return self.last


class AlphaBetaFilter(Filter):
    """See https://en.wikipedia.org/wiki/Alpha_beta_filter"""
    def __init__(self, a=0.85, b=0.005, x=None, v=None, dt=0.1, keys=[]):
        """Create new `AlphaBetaFilter` object.

        :param a: Float or list or dict of alpha weights.
        :param b: Float or list or dict of beta weights.
        :param x: Float or list or dict of x starting values.
        :param v: Float or list or dict of v starting values.
        :param dt: Time step of the filter.
        :param keys: The list of keys if dict is provided to `_update()`.
        """
        self.dt = float(dt)
        self.keys = keys
        if len(self.keys) > 0:
            self.alpha = {}
            self.beta = {}
            self.x = None if x is None else {}
            self.v = None if v is None else {}
            i = 0
            for k in self.keys:
                self.alpha[k] = _get_float(a, i)
                self.beta[k] = _get_float(b, i)
                if x is not None:
                    self.x[k] = _get_float(x, i)
                if v is not None:
                    self.v[k] = _get_float(v, i)
                i += 1
        else:
            self.alpha = float(a)
            self.beta = float(b)
            self.x = None if x is None else float(x)
            self.v = None if v is None else float(v)

    def _update(self, curr):
        """Compute value of curr in AB filter fashion."""
        def f(c, a, b, x, v, dt):
            xk = x + (v * dt)
            vk = v
            rk = c - xk

            xk += a * rk
            vk += (b * rk) / dt
            return xk, vk
        if self.x is None:
            self.x = curr
        elif isinstance(curr, dict):
            if self.v is None:
                self.v = {}
                for k in self.keys:
                    self.v[k] = (curr[k] - self.x[k]) / self.dt
            for k in curr:
                if k in self.keys:
                    self.x[k], self.v[k] = f(
                        curr[k],
                        self.alpha[k],
                        self.beta[k],
                        self.x[k],
                        self.v[k],
                        self.dt)
                else:
                    self.x[k] = curr[k]
        else:
            if self.v is None:
                self.v = (curr - self.x) / self.dt
            self.x, self.v = f(
                curr, self.alpha, self.beta, self.x, self.v, self.dt)
        return self.x, self.v
